/*
 * This file is a part of the CaosDB Project.
 *
 * Copyright (C) 2021 Timm Fitschen <t.fitschen@indiscale.com>
 * Copyright (C) 2021 IndiScale GmbH <info@indiscale.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 */
#include "caosdb/connection.h"         // for Connection, ConnectionManager
#include "caosdb/entity.h"             // for Entity, Messages, Message
#include "caosdb/message_code.h"       // for ENTITY_DOES_NOT_EXIST, Messag...
#include "caosdb/status_code.h"        // for SUCCESS, StatusCode
#include "caosdb/transaction.h"        // for Entity, Transaction, UniqueRe...
#include "caosdb/transaction_status.h" // for TransactionStatus, StatusCode
#include "gtest/gtest-message.h"       // for Message
#include "gtest/gtest-test-part.h"     // for TestPartResult, SuiteApiResolver
#include "gtest/gtest_pred_impl.h"     // for Test, EXPECT_EQ, AssertionResult
#include <memory>                      // for unique_ptr, allocator, __shar...
#include <string>                      // for string

namespace caosdb::transaction {
using caosdb::entity::Entity;
using caosdb::entity::MessageCode;

class test_transaction : public ::testing::Test {
protected:
  void SetUp() override {}

  void TearDown() override {
    // TODO(tf): delete all created entities
  }
};

TEST(test_transaction, DISABLED_retrieve_manufacturer_by_id) {
  const auto &connection =
    caosdb::connection::ConnectionManager::GetDefaultConnection();

  const auto *id = "107";
  const auto *role = "RecordType";
  const auto *name = "Manufacturer";
  const auto *description = "A generic manufacturer of all kinds of products";
  const auto *version = "0bea8f7b17f0130fa5701a6c3849b9f8bfa0651b";

  auto transaction(connection->CreateTransaction());
  transaction->RetrieveById(id);
  transaction->Execute();

  const auto &result_set =
    dynamic_cast<const UniqueResult &>(transaction->GetResultSet());

  const auto &entity = result_set.GetEntity();
  EXPECT_EQ(id, entity.GetId());
  EXPECT_EQ(name, entity.GetName());
  EXPECT_EQ(role, entity.GetRole());
  EXPECT_EQ(description, entity.GetDescription());
  EXPECT_EQ(version, entity.GetVersionId());
}

TEST(test_transaction, retrieve_non_existing) {
  const auto &connection =
    caosdb::connection::ConnectionManager::GetDefaultConnection();

  auto transaction(connection->CreateTransaction());

  const auto *id = "non-existing-id";
  transaction->RetrieveById(id);
  transaction->ExecuteAsynchronously();

  auto status = transaction->WaitForIt();
  EXPECT_EQ(status.GetCode(), TransactionStatus::TRANSACTION_ERROR().GetCode());
  ASSERT_EQ(status.GetCode(), StatusCode::GENERIC_TRANSACTION_ERROR);

  const auto &result_set =
    dynamic_cast<const UniqueResult &>(transaction->GetResultSet());

  const auto &entity = result_set.GetEntity();
  EXPECT_EQ(id, entity.GetId());
  EXPECT_TRUE(entity.HasErrors());
  ASSERT_EQ(entity.GetErrors().Size(), 1);
  EXPECT_EQ(entity.GetErrors().At(0).GetCode(),
            MessageCode::ENTITY_DOES_NOT_EXIST);
}

TEST(test_transaction, insert_delete) {
  const auto &connection =
    caosdb::connection::ConnectionManager::GetDefaultConnection();

  auto insert_transaction(connection->CreateTransaction());

  Entity entity;
  entity.SetRole("RecordType");
  entity.SetName("RT1");
  insert_transaction->InsertEntity(&entity);
  insert_transaction->ExecuteAsynchronously();

  auto insert_status = insert_transaction->WaitForIt();

  ASSERT_TRUE(insert_status.IsTerminated());
  ASSERT_FALSE(insert_status.IsError());

  const auto &insert_result_set =
    dynamic_cast<const UniqueResult &>(insert_transaction->GetResultSet());

  const auto &new_entity = insert_result_set.GetEntity();
  EXPECT_FALSE(new_entity.GetId().empty());
  EXPECT_FALSE(new_entity.HasErrors());

  auto delete_transaction(connection->CreateTransaction());

  delete_transaction->DeleteById(new_entity.GetId());
  delete_transaction->ExecuteAsynchronously();

  auto delete_status = delete_transaction->WaitForIt();

  ASSERT_TRUE(delete_status.IsTerminated());
  ASSERT_FALSE(delete_status.IsError());

  const auto &delete_result_set =
    dynamic_cast<const UniqueResult &>(delete_transaction->GetResultSet());

  const auto &deleted_entity = delete_result_set.GetEntity();
  EXPECT_EQ(deleted_entity.GetId(), new_entity.GetId());
  EXPECT_FALSE(deleted_entity.HasErrors());
}

} // namespace caosdb::transaction
